#include "sht.h"

#include <math.h> 
#include <time.h>
#include <stdlib.h>

HashTable* sht_new(){
     /* ROLE         Allocate dynamicall a new HashTable data structure 
                     and initialize it. 
        RETURNS      Pointer to the newly allocated data structure
        PARAMETERS   n/a
     */
     HashTable* p = NULL; 
           
     p = malloc(sizeof(HashTable)); 
     if (!p){ 
        perror("Unable to allocate hash table"); 
        exit (EXIT_FAILURE); 
     }
           
     p->tablesize = _HASH_SIZE_; 
     p->primer    = _HASH_SIZE_PRIMER_; 
     p->keysize   = _HASH_KEY_STRL_;
     p->nbpairs   = 0; 
     p->buckets = malloc(sizeof(Map*) * p->tablesize); 
           
     if(!p->buckets){
                     perror("Unable to allocate buckets"); 
                     exit(EXIT_FAILURE);
     }
           
     int i; 
     for(i=0 ; i < p->tablesize ; i++)
             p->buckets[i] = NULL; 
     return p;
}



void sht_destroy(HashTable* pht){
     /* ROLE         Deallocate all parts of a HashTable data structure
        RETURNS      n/a
        PARAMETERS   Pointer to the HashTable to deallocate
     */
     
     
     // Detecting possible malfunctions
     if(!pht){              perror("Unable to deallocate NULL hashtable"); 
                            exit(EXIT_FAILURE);
     }
     if(!pht->buckets){     perror("Unable to deallocate NULL buckets"); 
                            exit(EXIT_FAILURE);
     }
          
     int n; 
     // For all buckets in the HashTable...
     for(n=0 ; n < pht->tablesize ; n++){
             if (pht->buckets[n]){ 
                if(pht->buckets[n]->key) free(pht->buckets[n]->key); 
                if(pht->buckets[n]->value) free(pht->buckets[n]->value);
                free(pht->buckets[n]);
             }
     }
}


int sht_size( HashTable* pht ){
     /* ROLE         Returns the # of elements stored in the HashTable
        RETURNS      See above
        PARAMETERS   Pointer to the HashTable
     */
     return pht->nbpairs; 
} 

int sht_isempty( HashTable* pht){
     /* ROLE         Returns a boolean value indicating whether the HashTable
                     has no elements stored.
        RETURNS      See above
        PARAMETERS   Pointer to the HashTable
     */
    return sht_size(pht) == 0;     
}

Map** sht_lookup( HashTable* pht, char* k){ 
     /* ROLE         Looks up the hashtable for a given key k
        RETURNS      Returns address of the bucket pointer which points to the 
                     Map data structure associated with key k. 
                     Returns NULL if value is associated with k. 
        PARAMETERS   pht - Pointer to the HashTable
                     k   - key to lookup
     */

     int hashvalue; 
     hashvalue = hashFunction(k);
     Map** found = &(pht->buckets[hashvalue]); 
     if(!found){   perror("sht_remove unable to lookup key"); 
                   exit(EXIT_FAILURE); 
     }
     return found;     
}

int sht_put( HashTable* pht, char* k, char* v){
     /* ROLE         Adds a pair {k,v} to the hashtable
        RETURNS      1   - insertion was successful 
                     0   - Key k was already associated to a value, no insertion
        PARAMETERS   pht - Pointer to the HashTable
                     k   - key for new pair of elements
                     v   - value 
     */
     
     #ifdef _DEBUG_MODE_ 
     printf("PUT\t {%s,%s}", k, v);
     #endif 
     
     Map** found = sht_lookup(pht, k); 
     
     if(*found){
                 // The key k is alraedy associated with an element
                 
                 #ifdef _DEBUG_MODE_
                 printf("\tPrevious Value Kept, No Insertions\n"); 
                 sht_debug_display(pht); printf("\n\n\n");
                 #endif
                 
                 return 0; 
     }
     
     // The key k is not associated with an element
     *found = malloc(sizeof(Map));
     if (!(*found)){
        perror("Unable to allocate Map Pair"); 
        exit(EXIT_FAILURE);
     }
                 
     // Make a copy of the key and value strings 
     (*found)->value = strdup(v); 
     (*found)->key   = strdup(k);  
     
     // We have now one more element stored in our hashtable
     pht->nbpairs++;
     
     #ifdef _DEBUG_MODE_
     printf("\n"); sht_debug_display(pht); printf("\n\n\n");
     #endif
     
     return 1;
}
 


int sht_remove ( HashTable* pht, char* k ){
     /* ROLE         Remove an element associated with key k from the hashtable
                     Doesn't handle Collisions
        RETURNS      1   - removal was successful 
                     0   - Key k was not already associated to a value, no rmv
        PARAMETERS   pht - Pointer to the HashTable
                     k   - key for new pair of elements
     */
     
     #ifdef _DEBUG_MODE_ 
     printf("REMOVE\t {%s,???}", k);
     #endif 
     
     Map** found = sht_lookup( pht, k); 
     
     if (*found){
                 // We have an element associated with key k 
                 
                 free(*found);
                 *found = NULL; 
                 
                 // one less element in the hash table
                 pht->nbpairs--;
                 
                 #ifdef _DEBUG_MODE_
                 printf("removal successful\n"); 
                 sht_debug_display(pht); printf("\n\n\n");
                 #endif 
                 
                 return 1;
     }else{
                 // No element is associated wiht key k 
                 
                 #ifdef _DEBUG_MODE_
                 printf("nothing to remove\n"); sht_debug_display(pht); printf("\n\n\n");
                 #endif
                 
                 return 0;
     }
}



char* sht_get( HashTable* pht, char* k){
     /* ROLE         Find an element associated with key k in the hash table
                     and returns the address of its strdup. 
        RETURNS      The address of a newly allocated copy of the value string 
                     NULL, if no element associated to key k was found
        PARAMETERS   pht - Pointer to the HashTable
                     k   - key for new pair of elements
     */     
     
     #ifdef _DEBUG_MODE_ 
     printf("GET\t {%s,???}", k);
     #endif 
     
     Map** found = sht_lookup( pht, k);

     #ifdef _DEBUG_MODE_
     if (*found)      printf("\tFound value \"%s\"\n", (*found)->value);
     else             printf("\tNothing found!!!\n");
     sht_debug_display(pht); printf("\n\n\n");
     #endif

     // We need not only a Map element but also a value stored in it
     if ((*found)&&((*found)->value))          return strdup((*found)->value);
     else                                      return NULL;
}




